package core.services;

import core.api.models.timemanager.tasks.repeatinginfo.DeadlineStrategy;
import core.api.models.timemanager.tasks.repeatinginfo.TaskRepeatDayInfo;
import core.api.models.timemanager.tasks.repeatinginfo.TaskRepeatWeekDayInfo;
import core.api.models.timemanager.tasks.repeatinginfo.TaskRepeatWeekInfo;
import core.entities.timemanager.*;
import core.repositories.timemanager.ScheduleRepository;
import core.repositories.timemanager.TaskRepository;
import core.repositories.timemanager.TaskSerieItemRepository;
import core.repositories.timemanager.TaskSeriesRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import util.Tupel;

import java.time.DayOfWeek;
import java.time.Duration;
import java.time.LocalDate;
import java.time.temporal.ChronoUnit;
import java.util.*;

@Service
public class TaskSeriesService {

    @Autowired private TaskRepository taskRepository;
    @Autowired private TaskSeriesRepository taskSeriesRepository;
    @Autowired private TaskSerieItemRepository taskSerieItemRepository;
    @Autowired private ScheduleRepository scheduleRepository;


    public ServiceExitCode createTaskSeries(TaskRepeatWeekInfo taskRepeatInfo) {
        HashMap<Task, Tupel<Integer, Integer>> offsetMap = calcWeeklyOffsetMap(taskRepeatInfo);

        TaskSerie taskSerie = new TaskSerie();
        List<Task> clonedTasks = new ArrayList<>();
        List<AbstractSchedule> clonedSchedules = new ArrayList<>();
        int weekDayIndex = 0;
        for(Map.Entry<Task, Tupel<Integer, Integer>> repeatingTaskInfo: offsetMap.entrySet()) {
            Task rootTask = repeatingTaskInfo.getKey();
            TaskSerieItem rootItem = new TaskSerieItem(rootTask, weekDayIndex);
            taskSerie.addItem(rootItem);

            int itemIndex = weekDayIndex +1;
            Tupel<Collection<Task>, Collection<AbstractSchedule>> repeatingResult = repeatTask(rootTask, taskRepeatInfo.getEndDate(), offsetMap, taskSerie, itemIndex);

            clonedTasks.addAll(repeatingResult.getValue00());
            clonedSchedules.addAll(repeatingResult.getValue01());

            weekDayIndex++;
        }


        taskSeriesRepository.save(taskSerie);
        taskRepository.saveAll(clonedTasks);
        taskSerieItemRepository.saveAll(taskSerie.getTasks());
        scheduleRepository.saveAll(clonedSchedules);

        return ServiceExitCode.OK;
    }

    private Tupel<Collection<Task>, Collection<AbstractSchedule>> repeatTask(Task rootTask, LocalDate endingDate, HashMap<Task, Tupel<Integer, Integer>> offsetMap,TaskSerie taskSerie, int itemIndex) {
        List<Task> clonedTasks = new ArrayList<>();
        List<AbstractSchedule> clonedSchedules = new ArrayList<>();

        LocalDate currentDate = rootTask.getStartDate().plusDays(offsetMap.get(rootTask).getValue00());
        while(currentDate.isBefore(endingDate)) {
            var cloneResult = rootTask.cloneTask();
            Task clonedRootTask = cloneResult.getValue00();
            clonedTasks.addAll(cloneResult.getValue01());
            clonedSchedules.addAll(cloneResult.getValue02());

            for(Task clonedTask : cloneResult.getValue01()) {
                TaskSerieItem item = new TaskSerieItem(clonedTask, itemIndex);
                taskSerie.addItem(item);
            }

            Tupel<Integer, Integer> offsetEntry = offsetMap.get(rootTask);
            clonedRootTask.shiftTask(offsetEntry.getValue00(), offsetEntry.getValue01());

            currentDate = currentDate.plusDays(offsetEntry.getValue00());
            itemIndex += offsetMap.size();
        }

        return new Tupel<>(clonedTasks, clonedSchedules);
    }

    private HashMap<Task, Tupel<Integer, Integer>> calcWeeklyOffsetMap(TaskRepeatWeekInfo weekInfo) throws NoSuchElementException {
        HashMap<Task, Tupel<Integer, Integer>> offsetMap = new HashMap<>();
        weekInfo.getWeekDayInfos().sort(Comparator.comparing(TaskRepeatWeekDayInfo::getDayOfWeek));
        for(int i=0; i<weekInfo.getWeekDayInfos().size(); i++) {
            Optional<Task> requestedTask = taskRepository.findById(weekInfo.getWeekDayInfos().get(i).getTaskID());
            if(requestedTask.isEmpty()) {
                throw new NoSuchElementException();
            } else {
                if(weekInfo.getDeadlineStrategy() == DeadlineStrategy.DEADLINE_EQUAL_START) {
                    int offset = weekInfo.getWeekDayInfos().get(i).getOffset()-1;
                    offsetMap.put(requestedTask.get(), new Tupel<>(offset, offset));
                } else {
                    int startingOffset = weekInfo.getWeekDayInfos().get(i).getOffset() -1;
                    int endingOffset = getEndingOffset(weekInfo, i, startingOffset);

                    offsetMap.put(requestedTask.get(), new Tupel<>(startingOffset, endingOffset));
                }
            }

        }
        return offsetMap;
    }

    private static int getEndingOffset(TaskRepeatWeekInfo weekInfo, int i, int startingOffset) {
        DayOfWeek nextWeekday;
        if(i == weekInfo.getWeekDayInfos().size()-1) {
            nextWeekday = weekInfo.getWeekDayInfos().get(0).getDayOfWeek();
        } else {
            nextWeekday = weekInfo.getWeekDayInfos().get(i +1).getDayOfWeek();
        }

        DayOfWeek currentWeekDay = weekInfo.getWeekDayInfos().get(i).getDayOfWeek();

        int endingOffset = startingOffset + Math.abs(nextWeekday.getValue() - currentWeekDay.getValue());
        return endingOffset;
    }

    public ServiceExitCode createTaskSeries(Task rootTask, TaskRepeatDayInfo taskRepeatInfo) {
        if(taskRepeatInfo.getDeadlineStrategy() == DeadlineStrategy.FIX_DEADLINE) {
            return ServiceExitCode.INVALID_PARAMETER;
        }

        List<Task> taskList = new ArrayList<>();
        List<AbstractSchedule> abstractSchedules = new ArrayList<>();
        TaskSerie taskSerie = new TaskSerie();
        TaskSerieItem rootItem = taskSerie.addTask(rootTask);
        rootTask.setTaskSerieItem(rootItem);



        taskSeriesRepository.save(taskSerie);
        taskRepository.saveAll(taskList);
        taskSerieItemRepository.saveAll(taskSerie.getTasks());
        scheduleRepository.saveAll(abstractSchedules);

        return ServiceExitCode.OK;
    }

    public List<AbstractSchedule> cloneSchedules(Task previousTask, Task nextTask) {
        long numberDays = ChronoUnit.DAYS.between(previousTask.getStartDate(), nextTask.getStartDate());

        List<AbstractSchedule> clonedSchedules = new ArrayList<>();
        for(AbstractSchedule abstractSchedule : previousTask.getBasicTaskSchedules()) {
            AbstractSchedule clonedSchedule = abstractSchedule.cloneSchedule();
            clonedSchedule.shiftSchedule(numberDays);

            clonedSchedules.add(clonedSchedule);
        }
        return clonedSchedules;
    }

    public void deleteTaskSeriesItem(Task task) {
        TaskSerieItem item = task.getTaskSerieItem();
        TaskSerie taskSerie = task.getTaskSerieItem().getTaskSerie();
        taskSerie.getTasks().remove(item);
        task.setTaskSerieItem(null);
        taskSerieItemRepository.delete(item);
        if(taskSerie.getTasks().isEmpty()) {
            for(TaskSerieItem taskSerieItem : taskSerie.getTasks()) {
                taskSerieItem.setTaskSerie(null);
            }
            taskSerie.getTasks().clear();
            taskSeriesRepository.delete(taskSerie);
        } else {
            repearIndexing(taskSerie);
        }
    }

    private void repearIndexing(TaskSerie taskSerie) {
        taskSerie.getTasks().sort(Comparator.comparingInt(TaskSerieItem::getSeriesIndex));
        List<TaskSerieItem> updatedItems = new ArrayList<>();
        int currentIndex = 1;
        for(TaskSerieItem taskSerieItem : taskSerie.getTasks()) {
            if(taskSerieItem.getSeriesIndex() != currentIndex) {
                taskSerieItem.setSeriesIndex(currentIndex);
                updatedItems.add(taskSerieItem);
            }

            currentIndex++;
        }

        taskSerieItemRepository.saveAll(updatedItems);
    }

    public void deleteTaskSerieByTaskgroup(Taskgroup taskgroup) {
        taskSerieItemRepository.deleteByTaskgroup(taskgroup);
        taskSeriesRepository.deleteUnreferenced();
    }

    public static String convertIndexToString(int index) {
        if(index < 10) {
            return "0" + index;
        } else {
            return String.valueOf(index);
        }
    }

    public static String insertNameIndex(int seriesIndex, String taskName) {
        return taskName.replaceAll("\\$\\{i}", convertIndexToString(seriesIndex));
    }
}
